from __future__ import print_function
import pytest

from ...neural.vec2vec import Model, ReLu, Softmax
from ...neural._classes.feed_forward import FeedForward
from ...neural._classes.batchnorm import BatchNorm
from ...neural._classes.elu import ELU
from ...neural._classes.maxout import Maxout
from ...neural.ops import NumpyOps
from ...api import clone, chain
from ...neural.util import to_categorical

from ...extra import datasets

@pytest.fixture(scope='module')
def mnist():
    train_data, dev_data, _ = datasets.mnist()
    train_X, train_y = NumpyOps().unzip(train_data)
    dev_X, dev_y = NumpyOps().unzip(dev_data)
    dev_y = to_categorical(dev_y, nb_classes=10)
    train_y = to_categorical(dev_y, nb_classes=10)
    return (train_X[:1000], train_y[:1000]), (dev_X, dev_y)


@pytest.fixture(scope='module')
def train_X(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return train_X


@pytest.fixture(scope='module')
def train_y(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return train_y


@pytest.fixture(scope='module')
def dev_X(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return dev_X


@pytest.fixture(scope='module')
def dev_y(mnist):
    (train_X, train_y), (dev_X, dev_y) = mnist
    return dev_y


def create_relu_softmax(depth, width):
    with Model.define_operators({'*': clone, '>>': chain}):
        model = ReLu(width, 784) >> Softmax(10, width)
    return model


def create_relu_batchnorm_softmax(depth, width):
    with Model.define_operators({'*': clone, '>>': chain}):
        model = BatchNorm(ReLu(width, 784)) >> Softmax(10, width)
    return model

def create_elu_maxout_softmax(depth, width):
    with Model.define_operators({'>>': chain}):
        model = ELU(width) >> Maxout(5) >> Softmax()
    return model

@pytest.fixture(params=[
    create_relu_softmax,
    create_elu_maxout_softmax,
    create_relu_batchnorm_softmax])
def create_model(request):
    return request.param


@pytest.mark.slow
@pytest.mark.parametrize(('depth', 'width', 'nb_epoch'), [(2, 8, 5)])
def test_small_end_to_end(depth, width, nb_epoch,
        create_model,
        train_X, train_y, dev_X, dev_y):
    model = create_model(depth, width)
    assert isinstance(model, FeedForward)
    losses = []
    with model.begin_training(train_X, train_y) as (trainer, optimizer):
        trainer.each_epoch.append(lambda: print(model.evaluate(dev_X, dev_y)))
        trainer.nb_epoch = nb_epoch
        for X, y in trainer.iterate(train_X, train_y):
            X = model.ops.asarray(X)
            y = model.ops.asarray(y)
            yh, backprop = model.begin_update(X, drop=trainer.dropout)
            backprop(yh-y, optimizer)
            loss = ((yh-y)**2).sum()
            losses.append(loss)
    assert losses[-1] < losses[0]



# cython: infer_types=True
# cython: cdivision=True

cimport cython
from libc.stdint cimport int32_t
from libc.string cimport memset, memcpy
from cymem.cymem cimport Pool


from .typedefs cimport weight_t

include "compile_time_constants.pxi"

IF USE_BLAS:
    from . cimport openblas

cdef extern from "math.h" nogil:
    weight_t exp(weight_t x)
    weight_t sqrt(weight_t x)


cdef class Matrix:
    cdef readonly Pool mem
    cdef weight_t* data
    cdef readonly int32_t nr_row
    cdef readonly int32_t nr_col


cdef class Vec:
    @staticmethod    
    cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:
        if n_classes == 2:
            return 0 if scores[0] > scores[1] else 1
        cdef int i
        cdef int best = 0
        cdef weight_t mode = scores[0]
        for i in range(1, n_classes):
            if scores[i] > mode:
                mode = scores[i]
                best = i
        return best

    @staticmethod
    cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:
        if nr == 0:
            return 0
        cdef int i
        cdef weight_t mode = x[0]
        for i in range(1, nr):
            if x[i] > mode:
                mode = x[i]
        return mode

    @staticmethod
    cdef inline weight_t sum(const weight_t* vec, int32_t nr) nogil:
        cdef int i
        cdef weight_t total = 0
        for i in range(nr):
            total += vec[i]
        return total

    @staticmethod
    cdef inline weight_t norm(const weight_t* vec, int32_t nr) nogil:
        cdef weight_t total = 0
        for i in range(nr):
            total += vec[i] ** 2
        return sqrt(total)

    @staticmethod
    cdef inline void add(weight_t* output, const weight_t* x,
            weight_t inc, int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        Vec.add_i(output, inc, nr)

    @staticmethod
    cdef inline void add_i(weight_t* vec, weight_t inc, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] += inc

    @staticmethod
    cdef inline void mul(weight_t* output, const weight_t* vec, weight_t scal,
            int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.mul_i(output, scal, nr)

    @staticmethod
    cdef inline void mul_i(weight_t* vec, weight_t scal, int32_t nr) nogil:
        cdef int i
        IF USE_BLAS:
            openblas.scale(vec, nr, scal)
        ELSE:
            for i in range(nr):
                vec[i] *= scal

    @staticmethod
    cdef inline void pow(weight_t* output, const weight_t* vec, weight_t scal,
            int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.pow_i(output, scal, nr)

    @staticmethod
    cdef inline void pow_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] **= scal

    @staticmethod
    @cython.cdivision(True)
    cdef inline void div(weight_t* output, const weight_t* vec, weight_t scal,
            int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.div_i(output, scal, nr)

    @staticmethod
    @cython.cdivision(True)
    cdef inline void div_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] /= scal

    @staticmethod
    cdef inline void exp(weight_t* output, const weight_t* vec, int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.exp_i(output, nr)

    @staticmethod
    cdef inline void exp_i(weight_t* vec, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] = exp(vec[i])

    @staticmethod
    cdef inline void reciprocal_i(weight_t* vec, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] = 1.0 / vec[i]

    @staticmethod
    cdef inline weight_t mean(const weight_t* X, int32_t nr_dim) nogil:
        cdef weight_t mean = 0.
        for x in X[:nr_dim]:
            mean += x
        return mean / nr_dim

    @staticmethod
    cdef inline weight_t variance(const weight_t* X, int32_t nr_dim) nogil:
        # See https://www.johndcook.com/blog/standard_deviation/
        cdef double m = X[0]
        cdef double v = 0.
        for i in range(1, nr_dim):
            diff = X[i]-m
            m += diff / (i+1)
            v += diff * (X[i] - m)
        return v / nr_dim


cdef class VecVec:
    @staticmethod
    cdef inline void add(weight_t* output,
                         const weight_t* x, 
                         const weight_t* y,
                         weight_t scale,
                         int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        VecVec.add_i(output, y, scale, nr)
   
    @staticmethod
    cdef inline void add_i(weight_t* x, 
                           const weight_t* y,
                           weight_t scale,
                           int32_t nr) nogil:
        cdef int i
        IF USE_BLAS:
            openblas.simple_axpy(x, nr,
                y, scale)
        ELSE:
            for i in range(nr):
                x[i] += y[i] * scale
    
    @staticmethod
    cdef inline void batch_add_i(weight_t* x, 
                           const weight_t* y,
                           weight_t scale,
                           int32_t nr, int32_t nr_batch) nogil:
        # For fixed x, matrix of y
        cdef int i, _
        for _ in range(nr_batch):
            VecVec.add_i(x,
                y, scale, nr)
            y += nr
 
    @staticmethod
    cdef inline void add_pow(weight_t* output,
            const weight_t* x, const weight_t* y, weight_t power, int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        VecVec.add_pow_i(output, y, power, nr)

   
    @staticmethod
    cdef inline void add_pow_i(weight_t* x, 
            const weight_t* y, weight_t power, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            x[i] += y[i] ** power
 
    @staticmethod
    cdef inline void mul(weight_t* output,
            const weight_t* x, const weight_t* y, int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        VecVec.mul_i(output, y, nr)
   
    @staticmethod
    cdef inline void mul_i(weight_t* x, 
            const weight_t* y, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            x[i] *= y[i]

    @staticmethod
    cdef inline weight_t dot(
            const weight_t* x, const weight_t* y, int32_t nr) nogil:
        cdef int i
        cdef weight_t total = 0
        for i in range(nr):
            total += x[i] * y[i]
        return total
 
    @staticmethod
    cdef inline int arg_max_if_true(
            const weight_t* scores, const int* is_valid, const int n_classes) nogil:
        cdef int i
        cdef int best = -1
        for i in range(n_classes):
            if is_valid[i] and (best == -1 or scores[i] > scores[best]):
                best = i
        return best

    @staticmethod
    cdef inline int arg_max_if_zero(
            const weight_t* scores, const weight_t* costs, const int n_classes) nogil:
        cdef int i
        cdef int best = -1
        for i in range(n_classes):
            if costs[i] == 0 and (best == -1 or scores[i] > scores[best]):
                best = i
        return best


cdef class Mat:
    @staticmethod
    cdef inline void mean_row(weight_t* Ex,
            const weight_t* mat, int32_t nr_row, int32_t nr_col) nogil:
        memset(Ex, 0, sizeof(Ex[0]) * nr_col)
        for i in range(nr_row):
            VecVec.add_i(Ex, &mat[i * nr_col], 1.0, nr_col)
        Vec.mul_i(Ex, 1.0 / nr_row, nr_col)

    @staticmethod
    cdef inline void var_row(weight_t* Vx,
            const weight_t* mat, const weight_t* Ex,
            int32_t nr_row, int32_t nr_col, weight_t eps) nogil:
        # From https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
        if nr_row == 0 or nr_col == 0:
            return
        cdef weight_t sum_, sum2
        for i in range(nr_col):
            sum_ = 0.0
            sum2 = 0.0
            for j in range(nr_row):
                x = mat[j * nr_col + i]
                sum2 += (x - Ex[i]) ** 2
                sum_ += x - Ex[i]
            Vx[i] = (sum2 - sum_**2 / nr_row) / nr_row
            Vx[i] += eps
 

cdef class MatVec:
    @staticmethod
    cdef inline void add_i(weight_t* mat,
            const weight_t* vec, weight_t scale, int32_t nr_row, int32_t nr_col) nogil:
        cdef int i
        for i in range(nr_row):
            VecVec.add_i(mat + (i * nr_col),
                vec, scale, nr_col)

    @staticmethod
    cdef inline void mul(weight_t* output,
                         const weight_t* mat,
                         const weight_t* vec,
                         int32_t nr_row, int32_t nr_col) nogil:
        memcpy(output, mat, sizeof(output[0]) * nr_row * nr_col)
        MatVec.mul_i(output, vec, nr_row, nr_col)

    @staticmethod
    cdef inline void mul_i(weight_t* mat,
                           const weight_t* vec,
                           int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        for i in range(nr_row):
            row = i * nr_col
            for col in range(nr_col):
                mat[row + col] *= vec[col]

    @staticmethod
    cdef inline void dot(weight_t* output,
                         const weight_t* mat,
                         const weight_t* vec,
                         int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        cdef double zero = 0.0
        cdef int nr_batch = 1
        IF USE_BLAS:
            openblas.simple_gemm(output,
                nr_batch, nr_row,
                vec, nr_batch, nr_col,
                mat, nr_row, nr_col, 0, 1)
        ELSE:
            for i in range(nr_row):
                row = i * nr_col
                for col in range(nr_col):
                    output[i] += mat[row + col] * vec[col]
    
    @staticmethod
    cdef inline void batch_dot(weight_t* output,
                         const weight_t* mat,
                         const weight_t* vec,
                         int32_t nr_row, int32_t nr_col, int32_t nr_batch) nogil:
        # Output dim: batch_size * nr_row
        # vec dim:    batch_size * nr_col
        # mat dim:    nr_row     * nr_col
        # batch_size must be M, because can't transpose C
        # so nr_row must be N
        # so nr_col must be K

        # vec:   M * K
        # mat.T: K * N
        # out:   M * N
        cdef int i, row, col
        cdef double one = 1.0
        IF USE_BLAS:
            openblas.simple_gemm(output,
                nr_batch, nr_row,
                vec, nr_batch, nr_col,
                mat, nr_row, nr_col, 0, 1)
        ELSE:
            for b in range(nr_batch):
                MatVec.dot(output,
                    mat, vec, nr_row, nr_col)
                output += nr_row
                vec += nr_col

    @staticmethod
    cdef inline void T_dot(weight_t* output,
                             const weight_t* mat,
                             const weight_t* vec,
                             int32_t nr_row,
                             int32_t nr_col) nogil:
        cdef int i, row, col
        cdef double zero = 0.0
        cdef double one = 1.0
        cdef int nr_batch = 1
        IF USE_BLAS:
            openblas.simple_gemm(output,
                nr_batch, nr_row,
                vec, nr_batch, nr_col,
                mat, nr_row, nr_col, 0, 1)
        ELSE:
            for row in range(nr_row):
                for col in range(nr_col):
                    output[col] += vec[row] * mat[(row * nr_col) + col]

    @staticmethod
    cdef inline void batch_T_dot(weight_t* output,
                             const weight_t* mat,
                             const weight_t* vec,
                             int32_t nr_row,
                             int32_t nr_col,
                             int32_t nr_batch) nogil:
        cdef int _
        cdef double one = 1.0
        IF USE_BLAS:
            # output is (nr_batch, nr_col)
            # mat is (nr_row, nr_col)
            # vec is (nr_batch, nr_row)
            # Output defined as (M, N)
            # So
            # nr_batch = M
            # nr_col = N
            # nr_row = K
            #
            # vec:  M * K
            # mat:  K * N
            # out:  M * N
            openblas.simple_gemm(output,
                nr_batch, nr_col,
                vec, nr_batch, nr_row,
                mat, nr_row, nr_col, 0, 0)
        ELSE:
            for _ in range(nr_batch):
                MatVec.T_dot(output,
                    mat, vec, nr_row, nr_col)
                output += nr_col
                vec += nr_row


cdef class MatMat:
    @staticmethod
    cdef inline void add(weight_t* output,
                         const weight_t* x,
                         const weight_t* y,
                         int32_t nr_row, int32_t nr_col) nogil:
        memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)
        MatMat.add_i(output, y, nr_row, nr_col)

    @staticmethod
    cdef inline void add_i(weight_t* x,
                           const weight_t* y,
                           int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        for i in range(nr_row):
            row = i * nr_col
            for col in range(nr_col):
                x[row + col] += y[row + col]

    @staticmethod
    cdef inline void mul(weight_t* output,
                         const weight_t* x,
                         const weight_t* y,
                         int32_t nr_row, int32_t nr_col) nogil:
        memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)
        MatMat.mul_i(output, y, nr_row, nr_col)

    @staticmethod
    cdef inline void mul_i(weight_t* x,
                           const weight_t* y,
                           int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        for i in range(nr_row):
            row = i * nr_col
            for col in range(nr_col):
                x[row + col] *= y[row + col]

    @staticmethod 
    cdef inline void add_outer_i(weight_t* mat,
                                 const weight_t* x,
                                 const weight_t* y,
                                 int32_t nr_row,
                                 int32_t nr_col) nogil:
        cdef int i, j, row
        cdef double one = 1.0
        cdef float alpha = 1.0
        IF USE_BLAS:
            openblas.simple_ger(mat, nr_row, nr_col,
                x, nr_row, y, nr_col)
        ELSE:
            for i in range(nr_row):
                row = i * nr_col
                for j in range(nr_col):
                    mat[row + j] += x[i] * y[j]

    @staticmethod 
    cdef inline void batch_add_outer_i(weight_t* output,
                                 const weight_t* x,
                                 const weight_t* y,
                                 int32_t nr_row,
                                 int32_t nr_col,
                                 int32_t nr_batch) nogil:
        # Output dim: nr_row * nr_col
        # x dim:    batch_size * nr_row
        # y dim:    batch_size * nr_col
        # 
        # Output is M*N (can't transpose)
        # nr_row = M
        # nr_col = N
        # batch_size = K

        # x.T:  M * K
        # y:    K * N
        # out:  M * N
        cdef double one = 1.0
        IF USE_BLAS:
            openblas.simple_gemm(output,
                nr_row, nr_col,
                x, nr_batch, nr_row,
                y, nr_row, nr_col, 0, 0)
        ELSE:
            for _ in range(nr_batch):
                for i in range(nr_row):
                    row = i * nr_col
                    for j in range(nr_col):
                        output[row + j] += x[i] * y[j]
                x += nr_row
                y += nr_col



from __future__ import division, unicode_literals
from numpy import prod
import numpy
import contextlib
import msgpack
import msgpack_numpy
from collections import OrderedDict

msgpack_numpy.patch()

from .. import util
from ..train import Trainer
from ..ops import NumpyOps, CupyOps
from ..mem import Memory
from ..util import get_ops, copy_array, ensure_path
from ... import check
from ... import describe
from ...check import equal_length, has_shape, is_sequence, is_float, is_array


class Model(object):
    '''Model base class.'''
    name = 'model'
    id = 0
    lsuv = False
    ops = NumpyOps()
    Ops = NumpyOps
    Trainer = Trainer
    drop_factor = 1.0
    descriptions = []
    on_data_hooks = []
    on_init_hooks = [] # Use this to add layers
    _operators = {}

    @classmethod
    @contextlib.contextmanager
    def define_operators(cls, operators):
        '''Bind operators to specified functions for the scope of the context:

        Example
        -------

            model = Model()
            other = Model()
            with Model.define_operators({"+": lambda self, other: "plus"}):
                print(model + other)
                # "plus"
            print(model + other)
            # Raises TypeError --- binding limited to scope of with block.
        '''
        old_ops = dict(cls._operators)
        for op, func in operators.items():
            cls._operators[op] = func
        yield
        cls._operators = old_ops

    @classmethod
    @contextlib.contextmanager
    def use_device(cls, device):
        '''Change the device to execute on for the scope of the block.'''
        if device == cls.ops.device:
            yield
        else:
            curr_Ops, curr_ops = (cls.Ops, cls.ops)
            cls.Ops = get_ops(device)
            cls.ops = cls.Ops()
            yield
            cls.Ops = curr_Ops
            cls.ops = curr_ops

    @property
    def input_shape(self):
        raise NotImplementedError

    @property
    def output_shape(self):
        raise NotImplementedError

    def __init__(self, *args, **kwargs):
        self.name = self.__class__.name
        self.ops = self.Ops()
        kwargs = self._update_defaults(args, kwargs)
        self._mem = Memory(self.ops)
        self._dims = {}
        if not hasattr(self, '_layers'):
            self._layers = []
        self.descriptions = dict(self.descriptions)
        self.on_init_hooks = list(self.on_init_hooks)
        self.on_data_hooks = list(self.on_data_hooks)

        for attr, install in self.descriptions.items():
            install(attr, self)
        for hook in self.on_init_hooks:
            hook(self, *args, **kwargs)
        self.set_id()

    def _update_defaults(self, args, kwargs):
        new_kwargs = {}
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                new_kwargs[key] = value
        return new_kwargs

    def set_id(self):
        Model.id += 1
        self.id = Model.id
        for layer in self._layers:
            layer.set_id()

    #@check.args(equal_length)
    @check.arg(1, is_sequence)
    def begin_training(self, train_X, train_y=None, **trainer_cfg):
        for hook in self.on_data_hooks:
            hook(self, train_X, train_y)
        return self.Trainer(self, **trainer_cfg)

    @check.arg(2, is_float)
    @check.arg(1, has_shape(('nB', 'nI')))
    def begin_update(self, X, drop=0.0):
        raise NotImplementedError

    def predict(self, X):
        y, _ = self.begin_update(X, drop=None)
        return y

    def predict_one(self, x):
        X = self.ops.expand_dims(x, axis=0)
        return self.predict(X)[0]

    @contextlib.contextmanager
    def use_params(self, params): # pragma: no cover
        backup = None
        weights = self._mem.weights
        if self.id in params:
            param = params[self.id]
            backup = weights.copy()
            copy_array(weights, param)
        if hasattr(self, '_layers'):
            contexts = [layer.use_params(params) for layer in self._layers]
            for context in contexts:
                next(context.gen)
        yield
        if backup is not None:
            copy_array(self._mem.weights, backup)
        for i, context in enumerate(contexts):
            # This is ridiculous, but apparently it's what you
            # have to do to make this work across Python 2/3?
            try:
                next(context.gen)
            except StopIteration:
                pass

    def __call__(self, x):
        '''
        x
            Must match expected type
            Must match expected shape
        '''
        return self.predict(x)

    def pipe(self, stream, batch_size=128):
        for batch in util.minibatch(stream, batch_size):
            ys = self.predict(batch)
            for y in ys:
                yield y

    def update(self, stream, batch_size=1000):
        for X, y in util.minibatch(stream, batch_size=batch_size):
            output, finish_update = self.begin_update(X)
            gradient = finish_update(y)
            yield gradient

    def to_gpu(self, device_num):
        import cupy.cuda.device
        device = cupy.cuda.device.Device(device_num)
        device.use()
        queue = [self]
        for layer in queue:
            layer.ops = CupyOps()
            layer.Ops = CupyOps
            if hasattr(layer, u'_mem'):
                layer._mem._mem = self.ops.xp.asarray(layer._mem._mem)
                layer._mem.ops = layer.ops
            if hasattr(layer, u'_layers'):
                queue.extend(layer._layers)
        return device

    def to_cpu(self):
        queue = [self]
        for layer in queue:
            layer.ops = NumpyOps()
            layer.Ops = NumpyOps
            if hasattr(layer, u'_mem'):
                if hasattr(layer._mem._mem, 'get'):
                    layer._mem._mem = layer._mem._mem.get()
                layer._mem.ops = layer.ops
            if hasattr(layer, u'_layers'):
                queue.extend(layer._layers)

    def evaluate(self, X, y):
        '''
        x
            Must match expected type
            Must match expected shape
        y
            Must match expected type
        '''
        scores = self.ops.flatten(list(self.pipe(X)))
        if not hasattr(y, 'shape'):
            y = self.ops.flatten(y)
        scores = scores.reshape(y.shape)
        if len(scores.shape) == 1:
            correct = ((scores >= 0.5) == (y >= 0.5)).sum()
        else:
            correct = (scores.argmax(axis=1) == y.argmax(axis=1)).sum()
        return correct / y.shape[0]

    def evaluate_logloss(self, X, y, minimum=None, maximum=None):
        yh = self.ops.xp.vstack(self.pipe(X))
        yh = yh.reshape(y.shape)
        if minimum is not None:
            yh = self.ops.xp.maximum(yh, minimum)
        if maximum is not None:
            yh = self.ops.xp.minimum(yh, maximum)
        assert len(yh.shape) == 1
        losses = -y * self.ops.xp.log(yh + 1e-8) - (1-y) * self.ops.xp.log((1-yh)+1e-8)
        return losses.mean()

    @check.operator_is_defined('+')
    def __add__(self, other):
        '''Apply the function bound to the '+' operator.'''
        return self._operators['+'](self, other)

    @check.operator_is_defined('-')
    def __sub__(self, other):
        '''Apply the function bound to the '-' operator.'''
        return self._operators['-'](self, other)

    @check.operator_is_defined('*')
    def __mul__(self, other):
        '''Apply the function bound to the '*' operator.'''
        return self._operators['*'](self, other)

    @check.operator_is_defined('@')
    def __matmul__(self, other):
        '''Apply the function bound to the '@' operator.'''
        return self._operators['@'](self, other)

    @check.operator_is_defined('/')
    def __div__(self, other):
        '''Apply the function bound to the '/' operator.'''
        return self._operators['/'](self, other)

    @check.operator_is_defined('/')
    def __truediv__(self, other): # pragma: no cover
        '''Apply the function bound to the '/' operator.'''
        return self._operators['/'](self, other)

    @check.operator_is_defined('//')
    def __floordiv__(self, other):
        '''Apply the function bound to the '//' operator.'''
        return self._operators['//'](self, other)

    @check.operator_is_defined('%')
    def __mod__(self, other):
        '''Apply the function bound to the '%' operator.'''
        return self._operators['%'](self, other)

    @check.operator_is_defined('**')
    def __pow__(self, other, modulo=None):
        '''Apply the function bound to the '**' operator.'''
        return self._operators['**'](self, other)

    @check.operator_is_defined('<<')
    def __lshift__(self, other):
        '''Apply the function bound to the '<<' operator.'''
        return self._operators['<<'](self, other)

    @check.operator_is_defined('>>')
    def __rshift__(self, other):
        '''Apply the function bound to the '>>' operator.'''
        return self._operators['>>'](self, other)

    @check.operator_is_defined('&')
    def __and__(self, other):
        '''Apply the function bound to the '&' operator.'''
        return self._operators['&'](self, other)

    @check.operator_is_defined('^')
    def __xor__(self, other):
        '''Apply the function bound to the '^' operator.'''
        return self._operators['^'](self, other)

    @check.operator_is_defined('|')
    def __or__(self, other):
        '''Apply the function bound to the '|' operator.'''
        return self._operators['|'](self, other)

    def to_bytes(self):
        weights = []
        queue = [self]
        i = 0
        for layer in queue:
            if hasattr(layer, u'_mem'):
                weights.append(OrderedDict((
                    (b'dims', OrderedDict(sorted(layer._dims.items()))),
                    (b'params', []))))
                if hasattr(layer, u'seed'):
                    weights[-1][b'seed'] = layer.seed

                offsets = sorted(layer._mem._offsets.items())
                for (id_, name), (start, row, shape) in offsets:
                    if row == 1:
                        continue
                    param = layer._mem.get((id_, name))
                    if not isinstance(layer._mem.weights, numpy.ndarray):
                        param = param.get()
                    weights[-1][b'params'].append(
                        OrderedDict((
                            (b'name', name),
                            (b'offset', start),
                            (b'shape', shape),
                            (b'value', param),
                        ))
                    )
                i += 1
            if hasattr(layer, u'_layers'):
                queue.extend(layer._layers)
        return msgpack.dumps({b'weights': weights}, use_bin_type=True,
                             encoding='utf8')

    def from_bytes(self, bytes_data):
        data = msgpack.loads(bytes_data, encoding='utf8')
        weights = data[b'weights']
        queue = [self]
        i = 0
        for layer in queue:
            if hasattr(layer, '_mem'):
                if b'seed' in weights[i]:
                    layer.seed = weights[i][b'seed']
                for dim, value in weights[i][b'dims'].items():
                    if isinstance(dim, bytes):
                        dim = dim.decode('utf8')
                    setattr(layer, dim, value)
                for param in weights[i][b'params']:
                    name = param[b'name']
                    if isinstance(name, bytes):
                        name = name.decode('utf8')
                    dest = getattr(layer, name)
                    copy_array(dest, param[b'value'])
                i += 1
            if hasattr(layer, '_layers'):
                queue.extend(layer._layers)
        return self

    def to_disk(self, path):
        path = util.ensure_path(path)
        with path.open('wb') as file_:
            file_.write(self.to_bytes())

    def from_disk(self, path):
        path = util.ensure_path(path)
        with path.open('rb') as file_:
            bytes_data = file_.read()
        return self.from_bytes(bytes_data)



# cython: infer_types=True
# cython: cdivision=True
from libc.stdlib cimport calloc, free
from libc.string cimport memcpy, memset
from libc.math cimport sqrt
from cymem.cymem cimport Pool
from libcpp.vector cimport vector

from ._classes.maxout import Maxout
from ._classes.layernorm import LayerNorm
from ..linalg cimport Vec, Mat, VecVec, MatVec, MatMat
from .. cimport openblas

import numpy
cimport numpy as np

from ..api import wrap

ctypedef float real_t
ctypedef int dim_t


cdef extern from "stdlib.h":
    void* aligned_alloc(size_t alignment, size_t size) nogil


cdef class _Activations:
    cdef float** a
    cdef float** b
    cdef float** c
    cdef float** d
    cdef float** e
    cdef float** f
    cdef int** which
    cdef int* lengths
    cdef Pool mem
    cdef dim_t nX
    cdef dim_t nO
    cdef dim_t nP
    cdef dim_t nr_iter

    def __init__(self, lengths, dim_t nO, dim_t nP, dim_t nr_iter, gradient=False):
        self.mem = Pool()
        # Allocate buffers
        # Total e.g. nO=128, nP=3, nN=1000
        #   128*3*1000
        # + 128*3*1000
        # + 128*1000
        # + 128*1000
        # + 128*1000
        # = 384000 * 3 * 4 iterations
        # = approx 2mb
        nX = len(lengths)
        self.nX = nX
        self.nO = nO
        self.nP = nP
        self.nr_iter = nr_iter
        self.a = <float**>self.mem.alloc(nX, sizeof(float*))
        self.b = <float**>self.mem.alloc(nX, sizeof(float*))
        self.c = <float**>self.mem.alloc(nX, sizeof(float*))
        self.d = <float**>self.mem.alloc(nX, sizeof(float*))
        self.e = <float**>self.mem.alloc(nX, sizeof(float*))
        self.f = <float**>self.mem.alloc(nX, sizeof(float*))
        self.which = <int**>self.mem.alloc(nX, sizeof(int*))
        self.lengths = <int*>self.mem.alloc(nX, sizeof(int))
        cdef np.ndarray X
        cdef dim_t nN
        for i, nN in enumerate(lengths):
            self.a[i] = <float*>self.mem.alloc(nO*nN*nr_iter, sizeof(float))
            self.b[i] = <float*>self.mem.alloc(nO*3*nN*nr_iter, sizeof(float))
            self.c[i] = <float*>self.mem.alloc(nO*nP*nN*nr_iter, sizeof(float))
            self.d[i] = <float*>self.mem.alloc(nO*nN*nr_iter, sizeof(float))
            self.e[i] = <float*>self.mem.alloc(nO*nN*nr_iter, sizeof(float))
            self.f[i] = <float*>self.mem.alloc(nO*nN*nr_iter, sizeof(float))
            self.which[i] = <int*>self.mem.alloc(nO*nN*nr_iter, sizeof(int))
            self.lengths[i] = nN

    def set_inputs(self, Xs):
        cdef np.ndarray X
        for i, X in enumerate(Xs):
            memcpy(self.a[i], <float*>X.data, X.shape[0]*X.shape[1]*sizeof(float))

    def get_d_inputs(self):
        d_inputs = []
        cdef np.ndarray Y
        for i in range(self.nX):
            Y = numpy.zeros((self.lengths[i], self.nO), dtype='f')
            memcpy(<float*>Y.data,
                self.a[i], self.nO*self.lengths[i]*sizeof(float))
            d_inputs.append(Y)
        return d_inputs
 
    def set_d_outputs(self, dXs):
        cdef np.ndarray dX
        for i, dX in enumerate(dXs):
            memcpy(self.f[i], <float*>dX.data, dX.shape[0]*dX.shape[1]*sizeof(float))

    def get_outputs(self):
        outputs = []
        cdef np.ndarray Y
        for i in range(self.nX):
            Y = numpy.zeros((self.lengths[i], self.nO), dtype='f')
            memcpy(<float*>Y.data,
                self.a[i], self.nO*self.lengths[i]*sizeof(float))
            outputs.append(Y)
        return outputs
 

cdef class _Weights:
    cdef float* syn
    cdef float* bias
    cdef float* scale
    cdef float* shift
    def __init__(self, np.ndarray syn, np.ndarray bias,
                 np.ndarray scale, np.ndarray shift):
        self.syn = <float*>syn.data
        self.bias = <float*>bias.data
        self.scale = <float*>scale.data
        self.shift = <float*>shift.data


def MaxoutWindowEncoder(nr_unit, nr_iter):
    maxout = Maxout(nr_unit, nr_unit*3, pieces=3)
    normalize = LayerNorm(maxout)

    def mwe_fwd(Xs, drop=0.):
        if drop is not None and drop > 0:
            raise ValueError("MaxoutWindow encoder doesn't support dropout yet")
        return _mwe_fwd(nr_iter, maxout, normalize, Xs, drop=drop)

    model = wrap(mwe_fwd, normalize)
    model.maxout = maxout
    model.normalize = normalize
    return model


def _mwe_fwd(dim_t nr_iter, maxout, normalize, inputs, drop=0.):
    '''
    The function in the inner loop is:

    Given a:
    b, bp_b = window(a)
    c, bp_c = affine(b)
    d, bp_d = maxpool(c)
    e, bp_e = layernorm(d)
    f, bp_f = rescale(e)
    g = a + f
    return g, lambda dg: dg+bp_f(bp_e(bp_d(bp_c(bp_b(dg)))))

    In the backward pass we must compute:

    Given dg:
    df = dg 
    de = backprop_rescale(de)
    dd = backprop_layernorm(de)
    dc = backprop_maxpool(dd)
    db = backprop_affine(dc)
    da = backprop_window(db)
    Return dg+da
    '''
    ops = maxout.ops
    cdef dim_t nO = maxout.nO
    cdef dim_t nP = maxout.nP
    cdef dim_t nX = len(inputs)
    cdef np.ndarray lengths = ops.asarray([len(x) for x in inputs], dtype='i')
    cdef _Weights W = _Weights(maxout.W, maxout.b, normalize.G, normalize.b)
    cdef _Activations X = _Activations(lengths, nO, nP, nr_iter)
    X.set_inputs(inputs)

    cdef dim_t i
    for i in range(nX):
        mwe_forward(X.a[i], X.b[i], X.c[i], X.d[i], X.e[i], X.f[i], X.which[i],
            W.syn, W.bias, W.scale, W.shift, nO, nP, X.lengths[i], nr_iter)

    nonlocals = {}
    nonlocals['X'] = X
    nonlocals['lengths'] = lengths

    def mwe_bwd(d_outputs, sgd=None):
        cdef _Weights W = _Weights(maxout.W, maxout.b,
                                   normalize.G, normalize.b)
        cdef _Activations X = nonlocals['X']
        cdef np.ndarray lengths = nonlocals['lengths']
        cdef _Activations dX = _Activations(lengths, nO, nP, 1)
        dX.set_d_outputs(d_outputs) 

        cdef _Weights dW = _Weights(maxout.d_W, maxout.d_b,
                                    normalize.d_G, normalize.d_b)
        cdef dim_t i
        for i in range(nX):
            mwe_backward(
                dX.a[i], dX.b[i], dX.c[i], dX.d[i], dX.e[i], dX.f[i],
                dW.syn, dW.bias, dW.scale, dW.shift,
                    X.which[i], X.b[i], X.d[i], X.e[i],
                    W.syn, W.scale,
                    nO, nP, dX.lengths[i], nr_iter)
        if sgd is not None:
            sgd(maxout._mem.weights, maxout._mem.gradient, key=maxout.id)
            sgd(normalize._mem.weights, normalize._mem.gradient, key=normalize.id)
        return dX.get_d_inputs()
    return X.get_outputs(), mwe_bwd


cdef void mwe_forward(float* Xa, float* Xb, float* Xc, float* Xd, float* Xe, float* Xf, int* which,
        const float* syn, const float* bias, const float* scale, const float* shift,
        dim_t nO, dim_t nP, dim_t nN, dim_t nr_iter) nogil:
    # Now do the actual work
    for i in range(nr_iter):
        seq2col(Xb,
            Xa, 1, nO, nN)
        affine(Xc,
            Xb, syn, bias, nO*nP, nO*3, nN) 
        maxpool(Xd, which,
            Xc, nO, nP, nN)
        memcpy(Xe,
            Xd, nO*nN*sizeof(Xe[0]))
        layer_norm(Xe,
            nO, nN)
        memcpy(Xf,
            Xe, nO*nN*sizeof(float))
        rescale(Xf,
            scale, shift, nO, nN)
        VecVec.add_i(Xa,
            Xf, 1., nO*nN)
        Xb += nO*nP*nN
        memset(Xc, 0, nO*3*nN*sizeof(float))
        Xd += nO*nN
        Xe += nO*nN
        which += nO*nN


cdef void mwe_backward(
        float* dXa, float* dXb, float* dXc, float* dXd, float* dXe, float* dXf,
        float* dW_syn, float* dW_bias, float* dW_scale, float* dW_shift,
            const int* which, const float* Xb, const float* Xd, const float* Xe,
            const float* W_syn, const float* W_scale,
            dim_t nO, dim_t nP, dim_t nN, dim_t nr_iter) nogil:
    for i in range(nr_iter):
        which += nO*nN
        Xb += nO*3*nN
        Xd += nO*nN
        Xe += nO*nN
    for i in range(nr_iter-1, -1, -1):
        which -= nO*nN
        Xb -= nO*3*nN
        Xd -= nO*nN
        Xe -= nO*nN

        bwd_rescale(dXe, dW_scale, dW_shift,
            dXf, Xe, W_scale, nO, nN)
        bwd_layer_norm(dXd,
            dXe, Xd, nO, nN)
        bwd_maxpool(dXc,
            dXd, which, nO, nP, nN)
        bwd_affine(dXb, dW_syn, dW_bias,
            dXc, Xb, W_syn, nO*nP, nO*3, nN) 
        bwd_seq2col(dXa, 
            dXb, 1, nO, nN) 
        VecVec.add_i(dXa,
            dXf, 1., nN * nO)

        memcpy(dXf, dXa, nO*nN*sizeof(float))


cdef void seq2col(float* Xb,
        const float* Xa, dim_t nW, dim_t nI, dim_t nN) nogil:
    cdef dim_t nF = nW * 2 + 1
    Xb += nW * nI
    cdef dim_t i
    for i in range(nN-nW):
        memcpy(Xb,
            Xa, nI * (nW+1) * sizeof(Xb[0]))
        Xb += nI * (nW+1)
        memcpy(Xb,
            Xa, nI * nW * sizeof(Xb[0]))
        Xb += nI * nW
        Xa += nI
    memcpy(Xb,
        Xa, nI * nW * sizeof(Xb[0]))


cdef void bwd_seq2col(float* dXa,
        const float* dXb, dim_t nW, dim_t nI, dim_t nN) nogil:
    memset(dXa, 0, nI*nN*sizeof(float))
    # Here's what we're doing, if we had 2d indexing.
    #for i in range(B):
    #    d_seq[i] += d_cols[i-2, 4]
    #    d_seq[i] += d_cols[i-1, 3]
    #    d_seq[i] += d_cols[i+2, 0]
    #    d_seq[i] += d_cols[i+1, 1]
    #    d_seq[i] += d_cols[i, 2]
    nF = nW * 2 + 1
    for i in range(nN):
        seq_row = &dXa[i * nI]
        col_row = &dXb[i * nI * nF]
        for f in range(-nW, nW+1):
            if nN > (i+f) >= 0:
                feat = col_row + (f * nI)
                VecVec.add_i(seq_row, &feat[(f+nW) * nI], 1., nI)


cdef void affine(float* outputs,
        const float* inputs, const float* weights, const float* bias,
        dim_t nO, dim_t nI, dim_t nB) nogil:
    openblas.simple_gemm(outputs,
        nB, nO, inputs, nB, nI, weights, nO, nI, 0, 1)
    for i in range(nB):
        openblas.simple_axpy(&outputs[i*nO], nO,
            bias, 1.)


cdef void bwd_affine(float* d_inputs, float* dW, float* db,
        const float* d_outputs, const float* inputs, const float* weights,
        dim_t nO, dim_t nI, dim_t nB) nogil:
    memset(d_inputs, 0, nB*nI*sizeof(float))
    for i in range(nB):
        openblas.simple_axpy(db, nO,
            &d_outputs[i*nO], 1.)
    openblas.simple_gemm(dW, nO, nI,
        d_outputs, nB, nO, inputs, nB, nI, 1, 0)
    openblas.simple_gemm(d_inputs, nB, nI,
        d_outputs, nB, nO, weights, nO, nI, 0, 0)


cdef void maxpool(float* Xb, int* which,
        const float* Xa, dim_t nO, dim_t nP, dim_t nN) nogil:
    cdef int j
    for w in range(nN):
        for i in range(nO):
            j = Vec.arg_max(&Xa[w*nO*nP+i*nP], nP)
            Xb[w*nO+i] = Xa[w*nO*nP+i*nP+j]
            if which is not NULL:
                which[w*nO+i] = j


cdef void bwd_maxpool(float* dXa,
        const float* dXb, const int* which, dim_t nO, dim_t nP, dim_t nN) nogil:
    for b in range(nN):
        for o in range(nO):
            dXa[which[0]] = dXb[0]
            dXa += nP
            dXb += 1
            which += 1


cdef void layer_norm(real_t* X, dim_t nr_dim, dim_t nr_row) nogil:
    cdef double sqrt_var
    for i in range(nr_row):
        mu = Vec.mean(X, nr_dim)
        v = Vec.variance(X, nr_dim)
        if mu == 0. and v == 0:
            X += nr_dim
            continue
        sqrt_var = v ** -0.5
        for j in range(nr_dim):
            X[j] = (X[j] - mu) * sqrt_var
        X += nr_dim


cdef void bwd_layer_norm(real_t* dXa,
        const real_t* dXb, const real_t* Xa, dim_t nr_dim, dim_t nr_row) nogil:
    cdef double inv_var, inv_sqrt_var, sum_dXb_dist
    for i in range(nr_row):
        mu = Vec.mean(Xa, nr_dim)
        v = Vec.variance(Xa, nr_dim)
        inv_sqrt_var = v ** (-1./2)
        inv_var = 1. / v
        sum_dXb = Vec.sum(dXb, nr_dim)
        sum_dXb_dist = 0.
        for j in range(nr_dim):
            sum_dXb_dist += dXb[j] * (Xa[j]-mu)
        for j in range(nr_dim):
            dXa[j] = (
                nr_dim * dXb[j]
                - sum_dXb
                - (Xa[j]-mu) * inv_var * sum_dXb_dist
            )
            dXa[j] *= inv_sqrt_var
            dXa[j] /= nr_dim
        Xa += nr_dim
        dXa += nr_dim
        dXb += nr_dim
 

cdef void rescale(float* X,
        const float* G, const float* bias, int nO, int nB) nogil:
    for i in range(nB):
        VecVec.mul_i(X,
            G, nO)
        VecVec.add_i(X,
            bias, 1., nO)
        X += nO


cdef void bwd_rescale(float* d_down, float* dG, float* d_bias,
        const float* d_up, const float* down, const float* G, int nO, int nB) nogil:
    memcpy(d_down,
        d_up, nO*nB*sizeof(d_down[0]))
    for i in range(nB):
        VecVec.mul_i(d_down,
            G, nO)
        d_down += nO
    for i in range(nB):
        VecVec.add_i(d_bias,
            d_up, 1., nO)
    MatMat.batch_add_outer_i(dG,
        d_up, down, nO, 1, nB)



from .model import Model
from ...api import layerize
from .affine import Affine


class Residual(Model):
    def __init__(self, layer):
        Model.__init__(self)
        self._layers.append(layer)
        self.on_data_hooks.append(on_data)

    def __call__(self, X):
        Y = self._layers[0](X)
        if isinstance(X, list) or isinstance(X, tuple):
            return [X[i]+Y[i] for i in range(len(X))]
        else:
            return X + Y

    def begin_update(self, X, drop=0.):
        y, bp_y = self._layers[0].begin_update(X, drop=drop)
        if isinstance(X, list) or isinstance(X, tuple):
            output = [X[i]+y[i] for i in range(len(X))]
        else:
            output = X+y
        def residual_bwd(d_output, sgd=None):
            dX = bp_y(d_output, sgd)
            if isinstance(d_output, list) or isinstance(d_output, tuple):
                return [d_output[i]+dX[i] for i in range(len(d_output))]
            else:
                return d_output + dX
        return output, residual_bwd

def on_data(self, X, y=None):
    for layer in self._layers:
        for hook in layer.on_data_hooks:
            hook(layer, X, y)
        if hasattr(layer, 'W'):
            layer.W.fill(0)



import contextlib
from ..compat import BytesIO
from ..neural._classes.model import Model

try:
    import torch.autograd
    import torch.optim
    import torch
except ImportError:
    pass


class PyTorchWrapper(Model):
    '''Wrap a PyTorch model, so that it has the same API as Thinc models.
    To optimize the model, you'll need to create a PyTorch optimizer and call
    optimizer.step() after each batch --- see examples/wrap_pytorch.py
    '''
    def __init__(self, model):
        Model.__init__(self)
        self._model = model
        self._optimizer = None

    def begin_update(self, x_data, drop=0.):
        '''Return the output of the wrapped PyTorch model for the given input,
        along with a callback to handle the backward pass.
        '''
        x_var = torch.autograd.Variable(torch.Tensor(x_data),
                                        requires_grad=True)
        # Make prediction

        y_var = self._model(x_var)
        def backward_pytorch(dy_data, sgd=None):
            dy_var = torch.Tensor(dy_data)
            torch.autograd.backward((y_var,), grad_tensors=(dy_var,))
            dX = self.ops.asarray(x_var.grad.data)
            if sgd is not None:
                if self._optimizer is None:
                    self._optimizer = self._create_optimizer(sgd)
                self._optimizer.step()
                self._optimizer.zero_grad()
            return dX
        return self.ops.asarray(y_var.data), backward_pytorch

    def _create_optimizer(self, sgd):
        params = self._model.parameters()
        if sgd.b1 != 0 and sgd.b2 != 0:
            optimizer = torch.optim.Adam(params, lr=sgd.alpha, betas=(sgd.b1, sgd.b2))
        elif sgd.b2 == 0:
            optimizer = torch.optim.SGD(params, lr=sgd.alpha, momentum=sgd.b1)
        else:
            raise NotImplementedError
        return optimizer

    def to_disk(self, path):
        # TODO: Untested
        torch.save(self._model.state_dict(), str(path))

    def from_disk(self, path):
        # TODO: Untested
        self._model.load_state_dict(torch.load(path))

    def to_bytes(self):
        # TODO: Untested
        filelike = BytesIO()
        torch.save(self._model.state_dict(), filelike)
        filelike.seek(0)
        return filelike.getvalue()

    def from_bytes(self, data):
        # TODO: Untested
        filelike = BytesIO(data)
        filelike.seek(0)
        self._model.load_state_dict(torch.load(filelike))

    def to_gpu(self, device_num):
        self._model.cuda(device_num)

    def to_cpu(self):
        self._model.cpu()

    def resize_output(self, new_dim):
        #self.weight = nn.Parameter(F.pad(self.weight, ...)) # add classes
        #self.weight = nn.Parameter(F.pad(model.weight, ...)) # add classes
        raise NotImplementedError

    def resize_input(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def use_params(self, params): # pragma: no cover
        if self.id in params:
            backup = self.to_bytes()
            self.from_bytes(params[self.id])
        else:
            backup = None
        yield
        if backup is not None:
            self.from_bytes(backup)


class PyTorchWrapperRNN(PyTorchWrapper):
    '''Wrap a PyTorch RNN model
    '''
    def __call__(self, x_data, h_0=None):
        x_var = torch.autograd.Variable(torch.Tensor(x_data), requires_grad=False)
        # Make prediction
        out, h_n = self._model(x_var, h_0)
        return (self.ops.asarray(out.data), h_n)

    def begin_update(self, x_data, h_0=None, drop=0.):
        '''Return the output of the wrapped PyTorch model for the given input,
        along with a callback to handle the backward pass.
        '''
        x_var = torch.autograd.Variable(torch.Tensor(x_data),
                                        requires_grad=True)
        
        # Make prediction
        out, h_n = self._model(x_var, h_0)
        # Shapes will be:
        # out = seq_len, batch, hidden_size * num_directions
        # h_n = num_layers * num_directions, batch, hidden_size

        def backward_pytorch_rnn(d_data, sgd=None):
            dy_data, _ = d_data
            dout = torch.Tensor(dy_data)
            torch.autograd.backward((out,), grad_tensors=(dout,))
            dX = self.ops.asarray(x_var.grad.data)
            if sgd is not None:
                if self._optimizer is None:
                    self._optimizer = self._create_optimizer(sgd)
                self._optimizer.step()
                self._optimizer.zero_grad()
            return dX
        return (self.ops.asarray(out.data), h_n) , backward_pytorch_rnn

    def resize_output(self, new_dim):
        #self.weight = nn.Parameter(F.pad(self.weight, ...)) # add classes
        #self.weight = nn.Parameter(F.pad(model.weight, ...)) # add classes
        raise NotImplementedError

    def resize_input(self):
        raise NotImplementedError

    @contextlib.contextmanager
    def use_params(self, params): # pragma: no cover
        if self.id in params:
            backup = self.to_bytes()
            self.from_bytes(params[self.id])
        else:
            backup = None
        yield
        if backup is not None:
            self.from_bytes(backup)
